from __future__ import annotations

from poppusher.metadata import CountryMetadata

country: CountryMetadata = CountryMetadata(
    name_short_en="United Kingdom",
    name_official="The United Kingdom of Great Britain and Northern Ireland",
    iso3="GBR",
    iso2="GB",
    iso3166_2=None,
)

asset_prefix = "uk"
